/*
FILE:     CifFile.C
*/
/*
VERSION:  7.012
*/
/*
DATE:     4/3/2007
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2007 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/


/* 
  PURPOSE:    Base class for read/write cif files
*/

// VLAD - Carefully examine all methods that accept with, for cases where
// with is 0. This may happen if the value is empty !!

#include "GenString.h"
#include "CifString.h"
#include "regex.h"
#include "CifFile.h"


CifFile::CifFile(const eFileMode fileMode, const string& objFileName,
  const bool verbose, const StringCompare::eCompareType caseSense,
  const unsigned int maxLineLength, const string& nullValue) :
  TableFile(fileMode, objFileName, caseSense)
{

    Init();

    _maxCifLineLength = maxLineLength;
    _nullValue = nullValue;
    _verbose = verbose;

}


void CifFile::Init()
{

    _maxCifLineLength = STD_CIF_LINE_LENGTH;
    _nullValue = CifString::NullValue;
    _verbose = false;
    _smartPrint = false;
    _quotes = "\'";

}


void CifFile::SetQuoting(eQuoting quoting)
{

    if (quoting == eSINGLE)
    {
        _quotes = "\'";
    }
    else if (quoting == eDOUBLE)
    {
        _quotes = "\"";
    } 
    // VLAD - ERROR HANDLING
}


unsigned int CifFile::GetQuoting()
{

    if (_quotes == "\'")
        return(eSINGLE);
    else if (_quotes == "\"")
        return(eDOUBLE);
    else
        // VLAD - ERROR HANDLING
        return(eDOUBLE);
       
}


int CifFile::Write(const string& cifFileName, const bool sortTables,
  const bool writeEmptyTables)
{

    ofstream cifo(cifFileName.c_str(), ios::out | ios::trunc);

    int ret = Write(cifo, sortTables, writeEmptyTables);

    cifo.close();

    return(ret);

}


int CifFile::Write(ostream& cifo, const bool sortTables,
  const bool writeEmptyTables)
{

    vector<unsigned int> tablesIndices;

    if (sortTables)
    {
        GetSortedTablesIndices(tablesIndices);
    }
    else
    {
        GetTablesIndices(tablesIndices);
    }

    int ret = Write(cifo, tablesIndices, writeEmptyTables);

    return(ret);

}


int CifFile::Write(const string& cifFileName, const vector<string>& catOrder,
  const bool writeEmptyTables)
{

    ofstream cifo(cifFileName.c_str(), ios::out | ios::trunc);

    int ret = Write(cifo, catOrder, writeEmptyTables);

    cifo.close();

    return(ret);

}


int CifFile::Write(ostream& cifo, const vector<string>& catOrder,
  const bool writeEmptyTables)
{

    vector<bool> processedTables(GetTotalNumTables());
    vector<unsigned int> tables;
    unsigned int tableIndex = 0;

    // VLAD : Can this be improved by first searching the category and
    // then doing something about it? Well maybe this is not doable.
    for (unsigned int blockI = 0, numTables = 0; blockI < _blocks.size();
      blockI++)
    {
        for (unsigned int l = 0; l < catOrder.size(); l++)
        {
            tableIndex = _blocks[blockI]._tables.\
              find(catOrder[l]);

            if (tableIndex != _blocks[blockI]._tables.size())
            {
                tables.push_back(tableIndex);
                processedTables[tableIndex + numTables] = true;
            }
            else
            {
                processedTables[tableIndex + numTables] = false;
            }
        }
        numTables += _blocks[blockI]._tables.size();
    }

    for (unsigned int blockI = 0, numTables = 0; blockI < _blocks.size();
      ++blockI)
    {
        for (unsigned int tableI = 0; tableI < _blocks[blockI]._tables.size();
          ++tableI)
        {
            if (!processedTables[tableI + numTables])
            {
                tables.push_back(tableI);
            }
        }
        numTables += _blocks[blockI]._tables.size();
    }

    return(Write(cifo, tables, writeEmptyTables));

}


int CifFile::DataChecking(CifFile& ref)
{

    int ret1, ret2;

    ofstream log;

    log.open("Diagnostics.log", ios::out | ios::app);

    ret1 = MandatoryCategoryChecking(ref, log);
    ret2 = ItemsChecking(ref, log);

    log.close();

    if (ret1 == 0)
        return ret2;
    else
        return ret1;

}


int CifFile::_IsQuotableText(const string& itemValue)
{

  if (itemValue.empty())
      return(0);

  if (itemValue[0] == '_')
      return(1);

  for (unsigned int i = 0; i < itemValue.size(); i++)
  {
      if (itemValue[i] == ' ') 
          return(1);
      else if (itemValue[i] == '\n') 
          return(1);
      else if (itemValue[i] == '\'' || itemValue[i] == '\"') 
          return(1);
  }

  if (String::IsCiEqual(itemValue.substr(0, 5), "data_"))
      return(1);
  if (String::IsCiEqual(itemValue.substr(0, 5), "loop_"))
      return(1);
  if (String::IsCiEqual(itemValue.substr(0, 5), "save_"))
      return(1);

  return(0);

}


CifFile::eIdentType CifFile::_FindPrintType(const vector<string>& values)
{

    // If at least one value is a number, all should be right justified.
    for (unsigned int valuesI = 0; valuesI < values.size(); ++valuesI)
    {
        if (String::IsNumber(values[valuesI]))
        {
            return(eRIGHT);
        }
    }

    return(eLEFT);

}


void CifFile::_PrintItemIdent(ostream& cifo, unsigned int& linePos)
{

    string ItemNameIdent = "    ";
 
    cifo << ItemNameIdent;

    linePos = ItemNameIdent.size();

}


void CifFile::_PrintItemName(ostream& cifo, const string& category,
  const string& keyword, unsigned int& linePos)
{

    string cifItem;

    CifString::MakeCifItem(cifItem, category, keyword);

    cifo << cifItem;

    linePos += cifItem.size();

}


void CifFile::_PrintPostItemSeparator(ostream& cifo, unsigned int& linePos,
  const bool ident, const unsigned int numSpaces)
{

    const string Space = " ";

    for (unsigned int spaceI = 0; spaceI < numSpaces; spaceI++)
    {
        cifo << Space;
    }

    linePos += numSpaces * Space.size();

    if (!ident)
    {
        return;
    }

    unsigned int start = 36;

    while (linePos > start)
        start += 10;

    for (unsigned int i = 0; i < start - linePos; i++)
        cifo << Space;

    linePos = start;

}


int CifFile::_PrintItemValue(ostream& cifo, const string& itemValue,
  unsigned int& linePos, const eIdentType identType,
  const unsigned int width)
{

    bool multipleLine, multipleWord, embeddedQuotes, embeddedSingleQuotes,
      embeddedDoubleQuotes;
    string Ident;
    int N = 0;
    // VLAD: with has different meaning for eNONE and others.

    if ((identType == eNONE) && (width != 0))
    {
        Ident = "          ";
    }

    if (linePos == 0)
    {
        cifo << Ident;
        linePos = Ident.size();
    }
  
    if (itemValue.empty())
    {
        if (_maxCifLineLength <= linePos + 2)
        {
            cifo << endl;
            linePos = 0;
        }

        N = _nullValue.size();

        if (identType == eRIGHT)
        {
            if (linePos + width - N < _maxCifLineLength)
                for (unsigned int k = 0; k < width - N; k++)
                    cifo << " ";
                linePos += width - N;
        }

        cifo << _nullValue;
        linePos += 1;

        if (identType == eLEFT)
        {
            if (linePos != 0)
            {
                if (linePos + width - N < _maxCifLineLength)
                    for (unsigned int k = 0; k < width - N; k++)
                        cifo << " ";
                    linePos += width - N;
            }
        }

        cifo << " ";
        linePos += 1;

        return(1);
    }
  
    unsigned int str_len = itemValue.size();

    multipleLine = false;
    multipleWord = false;
    embeddedQuotes = false;
    embeddedSingleQuotes = false;
    embeddedDoubleQuotes = false;

    string multipleWordQuotes = _quotes;

    for (unsigned int i = 0; i < str_len; i++)
    {
        if (itemValue[i] == ' ')
            multipleWord = true;
        else if (itemValue[i] == '\n')
            multipleLine = true;
        else if (itemValue[i] == '\'')
        {
            embeddedSingleQuotes = true;
            embeddedQuotes = true;
            multipleWordQuotes = "\"";
        }
        else if (itemValue[i] == '\"')
        {
            embeddedDoubleQuotes = true;
            embeddedQuotes = true;
            multipleWordQuotes = "\'";
        }
    }

    if (itemValue[0] == '_')
        multipleWord = true;
    if (itemValue[0] == ';')
        multipleWord = true;

    if (String::IsCiEqual(itemValue.substr(0, 5), "data_"))
        multipleWord = true;
    if (String::IsCiEqual(itemValue.substr(0, 5), "loop_"))
        multipleWord = true;
    if (String::IsCiEqual(itemValue.substr(0, 5), "save_"))
        multipleWord = true;

    if (embeddedQuotes && multipleWord)
        multipleLine = true;
    if (embeddedQuotes)
        multipleWord = true;

    if (embeddedSingleQuotes && embeddedDoubleQuotes)
        multipleWordQuotes = _quotes;

    if (str_len >= _maxCifLineLength || multipleLine)
    {
        if (linePos != 0)
            cifo << endl;
        cifo << ";" << itemValue;
        cifo << endl;
        cifo << ";" << endl;
        linePos = 0;
    }
    else
    {
        if ((!multipleWord && str_len + 2 + linePos > _maxCifLineLength) ||
	  (multipleWord && str_len + 4 + linePos > _maxCifLineLength))
        {
            cifo << endl;
            linePos = 0;
            cifo << Ident;
            linePos += Ident.size();
        }

        string fullItemValue;

        if (multipleWord)
        {
            fullItemValue = multipleWordQuotes + itemValue + multipleWordQuotes;
        }
        else
        {
            fullItemValue = itemValue;
        }

        N = fullItemValue.size();

        if (identType == eRIGHT)
        {
            if (linePos + width - N < _maxCifLineLength)
                for (unsigned int k = 0; k < width - N; k++)
                    cifo << " ";
                linePos += width - N;
        }

        cifo << fullItemValue;
        linePos += N;

        if ((identType == eNONE) || (identType == eLEFT))
        {
            cifo << " ";
            linePos++;
            N++;
        }
    }
 
    if ((identType == eNONE) && (!Ident.empty()))
    {
        if (linePos > Ident.size())
        {
            linePos = linePos + width - N;
            for (int i = 0; i < (int)(width - N); i++)
                cifo << " ";
        }
    }

    if (identType == eLEFT)
    {
        if (linePos != 0)
        {
            if (linePos + width - (N - 1) < _maxCifLineLength)
                for (unsigned int k = 0; k < width - (N - 1); k++)
                    cifo << " ";
                linePos += width - (N - 1);
        }
    }

    if (identType == eRIGHT)
    {
        cifo << " ";
        linePos++;
        N++;
    }

    if (multipleWord || multipleLine || embeddedQuotes) 
        return(str_len + 2);
    else
        return(str_len);
 
}


int CifFile::_PrintItemNameInHeader(ostream& cifo, const string& itemValue,
  unsigned int& linePos, const eIdentType identType,
  const unsigned int width)
{

    bool multipleLine, multipleWord, embeddedQuotes, embeddedSingleQuotes,
      embeddedDoubleQuotes;
    string Ident;
    int N = 0;
    // VLAD: with has different meaning for eNONE and others.

    if ((identType == eNONE) && (width != 0))
    {
        Ident = "          ";
    }

    if (linePos == 0)
    {
        cifo << Ident;
        linePos = Ident.size();
    }
  
    if (itemValue.empty())
    {
        if (_maxCifLineLength <= linePos + 2)
        {
            cifo << endl << "# ";
            linePos = 2;
        }

        N = _nullValue.size();

        if (identType == eRIGHT)
        {
            if (linePos + width - N < _maxCifLineLength)
                for (unsigned int k = 0; k < width - N; k++)
                    cifo << " ";
                linePos += width - N;
        }

        cifo << _nullValue;
        linePos += 1;

        if (identType == eLEFT)
        {
            if (linePos != 0)
            {
                if (linePos + width - N < _maxCifLineLength)
                    for (unsigned int k = 0; k < width - N; k++)
                        cifo << " ";
                    linePos += width - N;
            }
        }

        cifo << " ";
        linePos += 1;

        return(1);
    }
  
    unsigned int str_len = itemValue.size();

    multipleLine = false;
    multipleWord = false;
    embeddedQuotes = false;
    embeddedSingleQuotes = false;
    embeddedDoubleQuotes = false;

    string multipleWordQuotes = _quotes;

    for (unsigned int i = 0; i < str_len; i++)
    {
        if (itemValue[i] == ' ')
            multipleWord = true;
        else if (itemValue[i] == '\n')
            multipleLine = true;
        else if (itemValue[i] == '\'')
        {
            embeddedSingleQuotes = true;
            embeddedQuotes = true;
            multipleWordQuotes = "\"";
        }
        else if (itemValue[i] == '\"')
        {
            embeddedDoubleQuotes = true;
            embeddedQuotes = true;
            multipleWordQuotes = "\'";
        }
    }

    if (itemValue[0] == '_')
        multipleWord = true;
    if (itemValue[0] == ';')
        multipleWord = true;

    if (String::IsCiEqual(itemValue.substr(0, 5), "data_"))
        multipleWord = true;
    if (String::IsCiEqual(itemValue.substr(0, 5), "loop_"))
        multipleWord = true;
    if (String::IsCiEqual(itemValue.substr(0, 5), "save_"))
        multipleWord = true;

    if (embeddedQuotes && multipleWord)
        multipleLine = true;
    if (embeddedQuotes)
        multipleWord = true;
    if (embeddedSingleQuotes && embeddedDoubleQuotes)
        multipleWordQuotes = _quotes;

    if (str_len >= _maxCifLineLength || multipleLine)
    {
        if (linePos != 0)
            cifo << endl;
        cifo << ";" << itemValue;
        cifo << endl;
        cifo << ";" << endl;
        linePos = 0;
    }
    else
    {
        if ((!multipleWord && str_len + 2 + linePos > _maxCifLineLength) ||
	  (multipleWord && str_len + 4 + linePos > _maxCifLineLength))
        {
            cifo << endl << "# ";
            linePos = 2;
            cifo << Ident;
            linePos += Ident.size();
        }

        string fullItemValue;

        if (multipleWord)
        {
            fullItemValue = multipleWordQuotes + itemValue + multipleWordQuotes;
        }
        else
        {
            fullItemValue = itemValue;
        }

        N = fullItemValue.size();

        if (identType == eRIGHT)
        {
            if (linePos + width - N < _maxCifLineLength)
                for (unsigned int k = 0; k < width - N; k++)
                    cifo << " ";
                linePos += width - N;
        }

        cifo << fullItemValue;
        linePos += N;

        if ((identType == eNONE) || (identType == eLEFT))
        {
            cifo << " ";
            linePos++;
            N++;
        }
    }
 
    if ((identType == eNONE) && (!Ident.empty()))
    {
        if (linePos > Ident.size())
        {
            linePos = linePos + width - N;
            for (int i = 0; i < (int)(width - N); i++)
                cifo << " ";
        }
    }

    if (identType == eLEFT)
    {
        if (linePos != 0)
        {
            if (linePos + width - (N - 1) < _maxCifLineLength)
                for (unsigned int k = 0; k < width - (N - 1); k++)
                    cifo << " ";
                linePos += width - (N - 1);
        }
    }

    if (identType == eRIGHT)
    {
        cifo << " ";
        linePos++;
        N++;
    }

    if (multipleWord || multipleLine || embeddedQuotes) 
        return(str_len + 2);
    else
        return(str_len);
 
}


int CifFile::Write(ostream& cifo, vector<unsigned int>& tables,
  const bool writeEmptyTables)
{

    ISTable* tableP = NULL;
 
    unsigned int numColumn, numRow, linePos, cwid, ilen, numPostItemSpaces;

    vector<string> colNames;
    vector<string> rowValues;
 
    vector<unsigned int> cwidth;
 
    if (_smartPrint)
    {
        numPostItemSpaces = SMART_PRINT_SPACING;
    }
    else
    {
        numPostItemSpaces = STD_PRINT_SPACING;
    }

 
    for (unsigned int blockI = 0, jt = 0;
      blockI < _blocks.size(); blockI++)
    {
        cifo << "data_" << _blocks[blockI].GetName() << endl;

        for (unsigned int tableI = 0; tableI < _blocks[blockI]._tables.size();
          ++tableI, ++jt)
        {
// VLAD - EXAMINE See this same loop in DICFileObjBase or DDLFileObjBase
// VLAD PERFORMANCE. The old BTreeObj code was doing a partial search
// on tids. It searched for "_0_ " in the tree. Even though the exact
// search was never found, it moved the pointer down the tree and reached
// the appropriate "_0_ blah" for example and then processed from that point
// onward the entries that are exact match. This kind of modified proprietory
// search is better than STL. We need to represent this subsearch using
// different STL structures to be able to have better performance than what
// we currently have.

            cifo << "# " << endl;

            tableP = _GetTablePtr(blockI, tables[jt]);

            numColumn = tableP->GetNumColumns();
            numRow = tableP->GetNumRows();

            if ((numRow == 0) && (!writeEmptyTables))
                continue;

            tableP->GetColumnNames(colNames);

            if (numRow <= 1)
            {
                // cwid is maxColumnNameSize
                unsigned int longestNameIndex = 0;

                cwid = 0;
                for (unsigned int i = 0; i < numColumn; i++)
                {
                    if (colNames[i].size() > cwid)
                    {
                          cwid = colNames[i].size();
                          longestNameIndex = i;
                    }
                }
 
                string longestCifItem;

                CifString::MakeCifItem(longestCifItem,
                  tableP->GetName(), colNames[longestNameIndex]);

                if (numRow == 0)
                {
                    for (unsigned int i = 0; i < numColumn; i++)
                    {
                        rowValues.push_back(_nullValue);
                    }
                }
                else
                {
                    tableP->GetRow(rowValues, 0);
                }

                for (unsigned int i = 0; i < numColumn; i++)
                {
                    linePos = 0;
                    _PrintItemName(cifo,
                      _blocks[blockI]._tables[tables[jt]].GetName(),
                      colNames[i], linePos);
                    _PrintPostItemSeparator(cifo, linePos, false,
                      numPostItemSpaces + colNames[longestNameIndex].size() -
                      colNames[i].size());

                    linePos = longestCifItem.size() + numPostItemSpaces - 1;

                    _PrintItemValue(cifo, rowValues[i], linePos);

                    if (linePos != 0)
                        cifo << endl;
                }
                rowValues.clear();
            }
            else
            {
                cifo << "loop_" << endl;

                for (unsigned int i = 0; i < numColumn; i++)
                {
                    linePos = 0;
                    _PrintItemName(cifo,
                      _blocks[blockI]._tables[tables[jt]].GetName(),
                      colNames[i], linePos);
                    _PrintPostItemSeparator(cifo, linePos);
                    cifo << endl;
                }

                unsigned int addSpace;

                for (unsigned int i = 0; i < numColumn; i++)
                {
                    if (_smartPrint)
                    {
                        // In smart printing, the column width is determined
                        // based on column values as well as column name
                        if (i == 0)
                        {
                            // "# "
                            addSpace = 2;
                        }
                        else
                        {
                            addSpace = 0;
                        }
                        cwidth.push_back(colNames[i].size() + addSpace);
                    }
                    else
                    { 
                        // Even if values are empty, they should occupy at
                        // least one character.
                        cwidth.push_back(1);
                    }
                }

                for (unsigned int l = 0; l < numRow; l++)
                {
                    tableP->GetRow(rowValues, l);
                    for (unsigned int i = 0; i < numColumn; i++)
                    {
                        ilen = rowValues[i].size();
                        if (_IsQuotableText(rowValues[i]))
                            ilen += 2;
                        if (ilen > cwidth[i])
                        {
                            if ((i == 0) && _smartPrint)
                                cwidth[i] = ilen + 2;
                            else
                                cwidth[i] = ilen;
                        }
                    }
                    rowValues.clear();
                }

                // In smart print mode, determine justification type for
                // each column.
                vector<eIdentType> colPrintType;
                vector<string> col;
                for (unsigned int colI = 0; colI < numColumn; ++colI)
                {
                    if (_smartPrint)
                    {
                        tableP->GetColumn(col, colNames[colI]);

                        colPrintType.push_back(_FindPrintType(col));

                        col.clear();
                    }
                    else
                    {
                        colPrintType.push_back(eLEFT);
                    }
                }

                for (unsigned int l = 0; l < numRow; l++)
                {
                    if (_smartPrint)
                    {
                        if ((l == 0) || ((l % HEADER_SPACING) == 0))
                        {
                            _PrintHeaderedItems(cifo, colNames, cwidth,
                              colPrintType);
                        }
                        // VLAD - PRINT - Add here printing the header
                        // every certain number of lines
                    }
                    linePos = 0;
                    tableP->GetRow(rowValues, l);
                    for (unsigned int i = 0; i < numColumn; i++)
                    {
                        _PrintItemValue(cifo, rowValues[i], linePos,
                          colPrintType[i], cwidth[i]);
                    }

                    if (linePos != 0)
                        cifo << endl;

                    rowValues.clear();
                }

                cwidth.clear();
            }
            colNames.clear();
        }
        cifo << "# " << endl;
    }

    return(1);

}

void CifFile::_PrintHeaderedItems(ostream& cifo,
  const vector<string>& colNames, const vector<unsigned int>& colWidths,
  const vector<eIdentType> colPrintType)
{

    unsigned int linePos = 0;
    unsigned int lessSpace = 0;

    cifo << "# ";

    linePos += 2;

    for (unsigned int colI = 0; colI < colNames.size(); ++colI)
    {
        if (colI == 0)
        {
            lessSpace = 2;
        }
        else
        {
            lessSpace = 0;
        }

        _PrintItemNameInHeader(cifo, colNames[colI], linePos,
          colPrintType[colI], colWidths[colI] - lessSpace);
    }

    cifo << endl;

}


/******* validation ******/

int CifFile::MandatoryCategoryChecking(CifFile& ref, ofstream& log)
{

    int ret = 1;
    vector<string> BlockNames;
    unsigned int numBlocks;
    ISTable* Category;
    vector<string> list;
    vector<string> target;
    vector<unsigned int> OutList;
    vector<string> refBlockNames;
    int refnumBlocks;
    string cell;


    ref.GetBlockNames(refBlockNames);
    refnumBlocks = refBlockNames.size();
    if (refnumBlocks > 1)
        log << "WARNING - dictionary has " << refnumBlocks <<
          " datablocks" << endl;

    Block& block = ref.GetBlock(refBlockNames[0]);

    Category = block.GetTablePtr("category");

    list.clear();
    list.push_back("mandatory_code");

    Category->CreateIndex("index0", list);

    target.clear();
    target.push_back("yes");

    Category->Search(OutList, target, list);
    if (!OutList.empty())
    {
        GetBlockNames(BlockNames);
        numBlocks = BlockNames.size();

        for (unsigned int i = 0; i < OutList.size(); i++)
        {
            // Do for all tables that are mandatory
            //Category->GetCell(cell, string("id"), OutList[i]);
            cell = (*Category)(OutList[i], "id");
            for (unsigned int j = 0; j < numBlocks; j++)
            {
                //do it for all datablocks
                if (!block.IsTablePresent(cell))
                {
                    log << "ERROR - table " << cell <<
                      " is mandatory, but is missing" << endl;
                    ret = 0;
	        }
            }
        }
        BlockNames.clear();
    }
    refBlockNames.clear();

    return ret;

}


int CifFile::ItemsChecking(CifFile& ref, ofstream& log)
{
  int ret = 1;
  int i,j,k,l,m;
  int numBlocks;
  int numTables;
  int numRows;
  int errCode;
  unsigned int iOut;
  int len;

  string ParentTblName;
  string keyword;
  string ParentName;
  string cell;
  string typeCode;

  ISTable * Category;
  ISTable * ParentTbl = NULL;
  ISTable * ParentChildTbl;

  vector<string> list;
  vector<string> target;
  vector<string> BlockNames;
  vector<string> ItemNames;
  vector<string> TableNames;
  vector<string> ParentList;  vector<string> ParentTarget;
  vector<unsigned int> OutList;
  
#define NS 10
  regex_t preg;
  regmatch_t pmatch[NS];
  int icmp = 0;
  int errCodeRegex;
  string patternString;
  bool keyNotFound;
  vector<pair<unsigned int, unsigned int> > duplRows;
  int errCodeP_CH;
  vector<string> nameList;

  ISTable * refItem;
  vector<string> refItemList;
  vector<string> refItemTarget;
  ISTable * ItemType,*ItemTypeL;
  vector<string> ItemTypeLList;
  vector<string> ItemTypeLTarget;
  ISTable * ItemRange;
  int errCodeRange;
  string* maxlist = NULL;
  string* minlist = NULL;
  int irange = 0;
  int matched;
  double dval, dmin, dmax;
  int ival, imin, imax;
  ISTable * ItemEnumeration;
  int errCodeEnumeration;
  string* enumlist = NULL;
  int ienum = 0;
  int ie;
  double de;
  ISTable * CategoryKey;
  vector<string> keyTarget;
  vector<string> keyList;


/***************************/


  /************** PARENT-CHILD **************/
  // For parent-chaild relationship checking we need table "item_linked"
  // index on "child_name" column, after search value from "parent_name" column
  Block& block = ref.GetBlock(ref.GetFirstBlockName());

  ParentChildTbl = block.GetTablePtr("item_linked");

  list.clear();
  list.push_back("child_name");
  // Creates an index on column child_name, table item_linked ddl/dictionary
  ParentChildTbl->CreateIndex("index0",list);


  /************** MANDATORY **************/
  // For mandatory checking we need tables "item" from ref
  // Table ref.Item - index on columns "catgory_id"+"mandatory_code",
  // after search value from "name" column
  // then checking for particular item in ctegory

  vector<string> inItemList;
  vector<string> inItemTarget;

  nameList.clear();
  nameList.push_back("name");

  refItem = block.GetTablePtr("item");

  refItemList.clear();
  refItemList.push_back("category_id");
  refItemList.push_back("mandatory_code");
  refItem->CreateIndex("index0",refItemList);

  /************** REGEX **************/
  // For regular expresson checking we need table "item_type"
  // index on column "name", after search value from "code" column
  // and table "item_type_list" (to find reg. exp.)
  // index on column "code, after search value from "construct" column

  ItemType = block.GetTablePtr("item_type");

  ItemTypeL = block.GetTablePtr("item_type_list");
  ItemTypeLList.clear();
  ItemTypeLList.push_back("code");

  
  /************** RANGE **************/
  // For range checking we need table "item_range"
  // index on column "name", after search value from "maximum" & "minimum" columns

  ItemRange = block.GetTablePtr("item_range");

  /************** ENUMERATION **************/
  // For enumeration checking we need table "item_enumeration"
  // index on column "name", after search value from "value" column
  ItemEnumeration = block.GetTablePtr("item_enumeration");

  /************** REDUNDANT ROWS - KEY **************/
  // For renduntant rows checking we need table "category_key"
  // index on column "id", after search value from "name" column
  keyList.clear();
  keyList.push_back("id");

  CategoryKey = block.GetTablePtr("category_key");

  vector<string> colNamesKey;


  GetBlockNames(BlockNames);
  numBlocks = BlockNames.size();
  
  for (i=0; i<numBlocks; i++) { //do for all datablocks
    Block& block = GetBlock(BlockNames[i]);
    block.GetTableNames(TableNames);
    numTables = TableNames.size();
    for (j=0; j<numTables; j++) { //do for all tables
      Category = block.GetTablePtr(TableNames[j]);
      /************** REDUNDANT ROWS - KEY **************/
      keyTarget.clear();
      keyTarget.push_back(TableNames[j]);
      CategoryKey->Search(OutList, keyTarget, keyList);
      if (!OutList.empty()) {
	colNamesKey.clear();
	for (k=0; k<(int)(OutList.size()); k++) {
	  //CategoryKey->GetCell(cell,string("name"),OutList[k]);
          cell = (*CategoryKey)(OutList[k], "name");
	  CifString::GetItemFromCifItem(keyword,cell);
	  colNamesKey.push_back(keyword);
	}

        keyNotFound = false;
        for (unsigned int i = 0; i < colNamesKey.size(); i++)
        {
            if (!Category->IsColumnPresent(colNamesKey[i]))
            {
                log << "ERROR - key column not found in table " <<
                    Category->GetName() << " - " << colNamesKey[i] << endl;
                keyNotFound = true;
                break;
            }
        }

        if (!keyNotFound)
        {
            Category->FindRedundantRows(colNamesKey, duplRows, 1);
            if (!duplRows.empty())
            {
                for (unsigned int rowI = 0; rowI < duplRows.size(); rowI++)
                {
                    log << "ERROR - duplicate row in table "<<
                      Category->GetName() << ". Row " << duplRows[rowI].first
                      << " is duplicate of row "<< duplRows[rowI].second <<
                      endl;
                }
            }
        }
      }
      /**************************************************/
      /*************** MANDATORY **************/
      refItemTarget.clear();
      refItemTarget.push_back(TableNames[j]);
      refItemTarget.push_back("yes");
      errCode = -1;
      refItem->Search(OutList, refItemTarget, refItemList);
      if (!OutList.empty()) {
        errCode = 0;
	for (k=0; k<(int)(OutList.size()); k++) {
	  //refItem->GetCell(cell,"name",OutList[k]);
          cell = (*refItem)(OutList[k], "name");
	  CifString::GetItemFromCifItem(keyword,cell);
	  if (!Category->IsColumnPresent(keyword)) {
	    log<<"ERROR - Mandatory item "<<cell<<" is not in datablock " << BlockNames[i] <<endl;
	    ret = 0;
	  }
	}
      }
      /****************************************/
      
      GetItemNames(ItemNames, *Category);
      for (k=0; k<(int)(ItemNames.size()); k++) { //all items

	target.clear();
	target.push_back(ItemNames[k]);

	/************** PARENT-CHILD **************/

	// Checking if item has a parent
        errCodeP_CH = -1;
	iOut = ParentChildTbl->FindFirst(target, list);
	if (iOut != ParentChildTbl->GetNumRows()) {
          errCodeP_CH = 0;

	  // if YES checking for "parent" table & value in "child" item
	  //ParentChildTbl->GetCell(ParentName,"parent_name",iOut);
          ParentName = (*ParentChildTbl)(iOut, "parent_name");
	  CifString::GetCategoryFromCifItem(ParentTblName,ParentName);
	  if (!block.IsTablePresent(ParentTblName)) {
	    //table for parent item does not exist
	    log <<"WARNING - Item "<<ItemNames[k]<<" has parent item "<< ParentName<<" but table "
		<< ParentTblName <<" is NOT present in the file"<<endl;
            // Set indication that the parent table does not exist. This will
            // prevent step 2 of parent-child relationship to be executed.
            errCodeP_CH = NOT_FOUND;
	  }
	  else {
            // Set indication that the parent table exists.
            errCodeP_CH = NO_TABLE_ERROR;
	    ParentTbl = block.GetTablePtr(ParentTblName);
	    CifString::GetItemFromCifItem(keyword,ParentName);
	    ParentList.clear();
	    ParentList.push_back(keyword);
#ifdef VLAD_DELETED
            // Exception is thrown here in cifexch test cases
            // See why is this logic wrong, for now leave the
            // index out. It will be created at search time.
	    if(!ParentTbl->IndexExists(keyword))
            {
	      ParentTbl->CreateIndex(keyword,ParentList);
	    }
#endif
	  }
	}

	/************** REGEX step 1**************/
	/* compile a regex */
        errCodeRegex = -1;
	iOut = ItemType->FindFirst(target, nameList);
	if (iOut != ItemType->GetNumRows()) {
          errCodeRegex = 0;
	  //ItemType->GetCell(typeCode,string("code"),iOut);
          typeCode = (*ItemType)(iOut, "code");
	  ItemTypeLTarget.clear();
	  ItemTypeLTarget.push_back(typeCode);
	  iOut = ItemTypeL->FindFirst(ItemTypeLTarget, ItemTypeLList);
	  //ItemTypeL->GetCell(cell,string("construct"),iOut);
          cell = (*ItemTypeL)(iOut, "construct");
          ConvertEscapedString(cell, patternString); 
	  icmp = rcsb_regcomp(&preg,patternString.c_str(),REG_EXTENDED);
          rcsb_regfree(&preg);
	}
	/****************************************/
	/************** RANGE step 1**************/
	/* finding min & max */
        errCodeRange = -1;
	ItemRange->Search(OutList, target, nameList);
	if (!OutList.empty()) {
          errCodeRange = 0;
          // VLAD: Why OutList->size() + 1 and not just OutList->size()?
	  maxlist = (string*) new string[OutList.size() + 1];
	  minlist = (string*) new string[OutList.size() + 1];
	  irange = OutList.size();
	  for (l=0;l<irange;l++) {
	    //ItemRange->GetCell(cell,string("maximum"),OutList[l]);
            cell = (*ItemRange)(OutList[l], "maximum");
	    maxlist[l] = cell;
	    //ItemRange->GetCell(cell,string("minimum"),OutList[l]);
            cell = (*ItemRange)(OutList[l], "minimum");
	    minlist[l] = cell;
	  }
	}
	/****************************************/
	/************ ENUMERATION step 1************/
        errCodeEnumeration = -1;
	ItemEnumeration->Search(OutList, target, nameList);
	if (!OutList.empty()) {
          errCodeEnumeration = 0;
	  enumlist = (string*) new string[OutList.size() + 1];
	  ienum = OutList.size();
	  for (l=0;l<ienum;l++) {
	    //ItemEnumeration->GetCell(cell,string("value"),OutList[l]);
            cell = (*ItemEnumeration)(OutList[l], "value");
	    enumlist[l] = cell;
	  }
	}
	/****************************************/
	// loop for all rows
	numRows = Category->GetNumRows();
	for (l=0; l<numRows ;l++){

          string localItem;
          CifString::GetItemFromCifItem(localItem, ItemNames[k]);
	  //Category->GetCell(cell,localItem,l);
          cell = (*Category)(l, localItem);
	  /*************** PARENT-CHILD step 2**************/
	  if (errCodeP_CH==0) {
	    if ((cell!=".") && (cell!="?") && (cell!="")) {
	      ParentTarget.clear();
	      ParentTarget.push_back(cell);
              try {
	      unsigned int resIndex = ParentTbl->FindFirst(ParentTarget,
                ParentList);
	      if (resIndex == ParentTbl->GetNumRows()) {
		log<<"ERROR - Value "<< cell <<" for "<<ItemNames[k]
		   <<" is NOT defined in parent item "<< ParentName <<endl;
		ret = 0;
	      }
              }
              catch (...)
              {
		log<<"ERROR - Value "<< cell <<" for "<<ItemNames[k]
		   <<" is NOT defined in parent item "<< ParentName <<endl;
		ret = 0;
              }
	    }
	  }
	  /****************************************/
	  
	  /*************** REGEX step 2**************/
	  if (errCodeRegex>=0) {
	    if (icmp==0 && (cell!=".") && (cell!="?") && (cell!="")) {
	      rcsb_regcomp(&preg,patternString.c_str(),REG_EXTENDED);
	      ret = rcsb_regexec(&preg,cell.c_str(),NS,pmatch,0);
              rcsb_regfree(&preg);
	      if (ret!=0) {
		log<<"ERROR - value "<<cell<<" for "<<ItemNames[k]<<" does not match the pattern (row "<<l<<")"<<endl;
		ret = 0;
	      }
	      else { 
		len=pmatch[0].rm_eo - pmatch[0].rm_so;
		if (len!=(int)(cell.size())&& (typeCode != "text")) { 
		  log<<"ERROR - value "<<cell<<" for "<<ItemNames[k]<<" does not match the pattern (row "<<l<<")"<<endl;
		  log<<"ERROR - for "<<ItemNames[k]<<" only "<<cell.substr(pmatch[0].rm_so)<<" is matched (row "<<l <<")"<<endl;
		  ret = 0;
		}
	      }
	    }
	    /****************************************/
	    /************** RANGE step 2**************/
	    //	    Range(cell,maxlist,minlist,
	    if (errCodeRange>=0&& (cell!=".") && (cell!="?") && (cell!="")) {
	  m=0;
	  matched = 0;
	  if (typeCode == "float") {
	    dval = String::StringToDouble(cell);
	    while (m<irange && !matched) {
	      if (((minlist[m] != "?") && (minlist[m] != ".")
                && (!minlist[m].empty())) && ((maxlist[m] != "?")
                && (maxlist[m] != ".") && (!maxlist[m].empty()))) {
		dmin = String::StringToDouble(minlist[m]);
		dmax = String::StringToDouble(maxlist[m]);
		if (dmin==dmax) {
		  if (dval==dmax) matched = 1;
		}
		else
		  if (dval > dmin && dval < dmax) matched = 1;
	      }
	      else {
		if ((minlist[m] != "?") && 
                  (minlist[m] != ".") && 
                  (!minlist[m].empty())) {
		  dmin = String::StringToDouble(minlist[m]);
		  if (dval > dmin) matched = 1;
		}
		else {
		  if ((maxlist[m] != "?") && 
                    (maxlist[m] != ".") && 
                    (!maxlist[m].empty())) {
		    dmax = String::StringToDouble(maxlist[m]);
		    if (dval < dmax) matched = 1;
		  }
		  else
		    matched = 1;
		}
	      }
	      m++;
	    }
	    if (matched==0) {
	      log<<"ERROR - for "<<ItemNames[k]<<" "<<cell<<" is NOT in a range (row "<<l <<")"<<endl;
	      ret = 0;
	    }
	  }
	  else {
	    m=0;
	    matched = 0;
	    if (typeCode == "int") {
	      ival = String::StringToInt(cell);
	      while (m<irange && !matched) {
	      if (((minlist[m] != "?") && (minlist[m] != ".") 
                && (!minlist[m].empty())) && ((maxlist[m] != "?")
                && (maxlist[m] != ".") && (!maxlist[m].empty())))
              {
		imin = String::StringToInt(minlist[m]);
		imax = String::StringToInt(maxlist[m]);
		if (imin==imax) {
		  if (ival==imin) matched = 1;
		}
		else
		  if (ival > imin && ival < imax) matched = 1;
	      }
	      else {
		if ((minlist[m] != "?") && 
                  (minlist[m] != ".") && 
                  (!minlist[m].empty())) {
		  imin = String::StringToInt(minlist[m]);
		  if (ival > imin) matched = 1;
		}
		else {
		  if ((maxlist[m] != "?") && 
                    (maxlist[m] != ".") && 
                    (!maxlist[m].empty())){
		    imax = String::StringToInt(maxlist[m]);
		    if (ival < imax) matched = 1;
		  }
		  else
		    matched = 1;
		}
	      }
		m++;
	      }
	      if (matched==0) {
	      log<<"ERROR - for "<<ItemNames[k]<<" "<<cell<<" is NOT in a range (row "<<l <<")"<<endl;
	      ret = 0;
	      }
	    }
	  }
	}
	  /****************************************/
	  /************ ENUMERATION step 2************/
	  if (errCodeEnumeration>=0 && (cell!=".") && (cell!="?") && (cell!="")) {
	    m=0;
	    matched = 0;
	    if (typeCode == "float") {
	      dval = String::StringToDouble(cell);
	      while (m<ienum && !matched) {
		de = String::StringToDouble(enumlist[m]);
		if (dval == de) matched = 1;
		m++;
	      }
	    }
	    else {
	      if (typeCode == "int") {
		ival = String::StringToInt(cell);
		while (m<ienum && !matched) {
		  ie = String::StringToInt(enumlist[m]);
		  if (ival == ie) matched = 1;
		  m++;
		}
	      }
	      else {
		while (m<ienum && !matched) {
		if (cell == enumlist[m]) matched = 1;
		m++;
		}
	      }
	    }
	    if (matched==0) {
	      log<<"ERROR - for "<<ItemNames[k]<<" "<<cell<<" is NOT defined in the enumeration list (row "<<l <<")"<<endl;
	      ret = 0;
	    }
	  }
	  /****************************************/
	  }
	} //for all rows
	
	if (errCodeRange>=0) {
	  if (maxlist) delete [] maxlist;
	  if (minlist) delete [] minlist;
	}
	if (errCodeEnumeration>=0) {
	  if (enumlist) delete [] enumlist;
	}
      
      } //for all items
      
      ItemNames.clear();
    }
    TableNames.clear();
  }
  
  BlockNames.clear();

  return ret;
}


// This method converts the string with embedded escape sequences into
// the real string. For example, a two character combination "\t" is
// converted into one character representing horizontal tab.
void CifFile::ConvertEscapedString(const string& inString,
  string& outString)
{

    outString.clear();

    outString.reserve(inString.size());

    // Go through the characters of the input string
    for (unsigned int posI = 0; posI < inString.size(); )
    {
        if (inString[posI] == '\\')
        {
            if (inString[posI + 1] == '\\')
            {
                  outString.push_back('\\');
                  posI += 2;
            }
            else if (inString[posI + 1] == 't')
            {
                  outString.push_back('\t');
                  posI += 2;
            }
            else if (inString[posI + 1] == 'n')
            {
                  outString.push_back('\n');
                  posI += 2;
            }
            else
            {
                  outString.push_back(inString[posI]);
                  posI++;
            }
        } // Current character is '\'
        else
        {
            outString.push_back(inString[posI]);
            posI++;
        } // Curent character is not '\'
    } // for (characters in the input string)

} // CifFile::ConvertEscapedString()


void CifFile::GetItemNames(vector<string>& itemNames, ISTable& isTable)
{

    vector<string> colNames;
    string cifItem;

    itemNames.clear();

    string tableName = isTable.GetName();

    isTable.GetColumnNames(colNames);

    itemNames.reserve(colNames.size());

    for (unsigned int i = 0; i < colNames.size(); i++)
    {
        CifString::MakeCifItem(cifItem, tableName, colNames[i]);

        itemNames.push_back(cifItem);
    }

}


